from collections import namedtuple

BERObject = namedtuple('BERObject', ['tag', 'content'])

class IncompleteBERError(ValueError):
	def __init__(self, expected_length=-1):
		super().__init__()
		self.expected_length = expected_length

def decode_ber(data):
	index = 0
	if len(data) < 2:
		raise IncompleteBERError(2)
	identifier = data[index]
	ber_class = identifier >> 6
	ber_constructed = bool(identifier & 0x20)
	ber_type = identifier & 0x1f
	index += 1
	if not data[index] & 0x80:
		length = data[index]
		index += 1
	elif data[index] == 0x80:
		raise ValueError('Indefinite form not implemented')
	elif data[index] == 0xff:
		raise ValueError('BER length invalid')
	else:
		num = data[index] & ~0x80
		index += 1
		if len(data) < index + num:
			raise IncompleteBERError(index + num)
		length = 0
		for octet in data[index:index + num]:
			length = length << 8 | octet
		index += num
	if len(data) < index + length:
		raise IncompleteBERError(index + length)
	ber_content = data[index: index + length]
	rest = data[index + length:]
	return BERObject((ber_class, ber_constructed, ber_type), ber_content), rest

def decode_ber_integer(data):
	if not data:
		return 0
	value = -1 if data[0] & 0x80 else 0
	for octet in data:
		value = value << 8 | octet
	return value

def encode_ber(obj):
	tag = (obj.tag[0] & 0b11) << 6 | (obj.tag[1] & 1) << 5 | (obj.tag[2] & 0b11111)
	length = len(obj.content)
	if length < 127:
		return bytes([tag, length]) + obj.content
	octets = []
	while length:
		octets.append(length & 0xff)
		length = length >> 8
	return bytes([tag, 0x80 | len(octets)]) + bytes(reversed(octets)) + obj.content

def encode_ber_integer(value):
	if value < 0 or value > 255:
		raise NotImplementedError('Encoding of integers greater than 255 is not implemented')
	return bytes([value])

class BERType:
	@classmethod
	def from_ber(cls, data):
		raise NotImplementedError()

	@classmethod
	def to_ber(cls, obj):
		raise NotImplementedError()

	def __bytes__(self):
		return type(self).to_ber(self)

class OctetString(BERType):
	ber_tag = (0, False, 4)

	@classmethod
	def from_ber(cls, data):
		obj, rest = decode_ber(data)
		if obj.tag != cls.ber_tag:
			raise ValueError('Expected tag %s but found %s'%(cls.ber_tag, obj.tag))
		return obj.content, rest

	@classmethod
	def to_ber(cls, obj):
		if not isinstance(obj, bytes):
			raise TypeError()
		return encode_ber(BERObject(cls.ber_tag, obj))

class Integer(BERType):
	ber_tag = (0, False, 2)

	@classmethod
	def from_ber(cls, data):
		obj, rest = decode_ber(data)
		if obj.tag != cls.ber_tag:
			raise ValueError()
		return decode_ber_integer(obj.content), rest

	@classmethod
	def to_ber(cls, obj):
		if not isinstance(obj, int):
			raise TypeError()
		return encode_ber(BERObject(cls.ber_tag, encode_ber_integer(obj)))

class Boolean(BERType):
	ber_tag = (0, False, 1)

	@classmethod
	def from_ber(cls, data):
		obj, rest = decode_ber(data)
		if obj.tag != cls.ber_tag:
			raise ValueError()
		return bool(decode_ber_integer(obj.content)), rest

	@classmethod
	def to_ber(cls, obj):
		if not isinstance(obj, bool):
			raise TypeError()
		content = b'\xff' if obj else b'\x00'
		return encode_ber(BERObject(cls.ber_tag, content))

class Set(BERType):
	ber_tag = (0, True, 17)
	set_type = OctetString

	@classmethod
	def from_ber(cls, data):
		setobj, rest = decode_ber(data)
		if setobj.tag != cls.ber_tag:
			raise ValueError()
		objs = []
		data = setobj.content
		while data:
			obj, data = cls.set_type.from_ber(data)
			objs.append(obj)
		return list(objs), rest

	@classmethod
	def to_ber(cls, obj):
		content = b''
		for item in obj:
			content += cls.set_type.to_ber(item)
		return encode_ber(BERObject(cls.ber_tag, content))

class SequenceOf(Set):
	ber_tag = (0, True, 16)

class Sequence(BERType):
	ber_tag = (0, True, 16)
	sequence_fields = [
		#(Type, attr_name, default_value, optional?),
	]

	def __init__(self, *args, **kwargs):
		for index, spec in enumerate(type(self).sequence_fields):
			field_type, name, default, optional = spec
			if index < len(args):
				value = args[index]
			elif name in kwargs:
				value = kwargs[name]
			else:
				value = default() if callable(default) else default
			setattr(self, name, value)

	def __repr__(self):
		args = []
		for field_type, name, default, optional in type(self).sequence_fields:
			args.append('%s=%s'%(name, repr(getattr(self, name))))
		return '<%s(%s)>'%(type(self).__name__, ', '.join(args))

	@classmethod
	def from_ber(cls, data):
		seqobj, rest = decode_ber(data)
		if seqobj.tag != cls.ber_tag:
			raise ValueError()
		args = []
		data = seqobj.content
		for field_type, name, default, optional in cls.sequence_fields:
			try:
				obj, data = field_type.from_ber(data)
				args.append(obj)
			except ValueError as e:
				if not optional:
					raise e
				args.append(None)
		return cls(*args), rest

	@classmethod
	def to_ber(cls, obj):
		if not isinstance(obj, cls):
			raise TypeError()
		content = b''
		for field_type, name, default, optional in cls.sequence_fields:
			if not optional or getattr(obj, name) is not None:
				content += field_type.to_ber(getattr(obj, name))
		return encode_ber(BERObject(cls.ber_tag, content))

class Choice(BERType):
	ber_tag = None

	@classmethod
	def from_ber(cls, data):
		obj, rest = decode_ber(data)
		for subcls in cls.__subclasses__():
			if subcls.ber_tag == obj.tag:
				return subcls.from_ber(data)
		return None, rest

	@classmethod
	def to_ber(cls, obj):
		for subcls in cls.__subclasses__():
			if isinstance(obj, subcls):
				return subcls.to_ber(obj)
		raise TypeError()

class Wrapper(BERType):
	ber_tag = None
	wrapped_attribute = None
	wrapped_type = None
	wrapped_default = None
	wrapped_clsattrs = {}

	def __init__(self, *args, **kwargs):
		cls = type(self)
		attribute = cls.wrapped_attribute
		if args:
			setattr(self, attribute, args[0])
		elif kwargs:
			setattr(self, attribute, kwargs[attribute])
		else:
			setattr(self, attribute, cls.wrapped_default() if callable(cls.wrapped_default) else cls.wrapped_default)

	def __repr__(self):
		return '<%s(%s)>'%(type(self).__name__, repr(getattr(self, type(self).wrapped_attribute)))

	@classmethod
	def from_ber(cls, data):
		class WrappedType(cls.wrapped_type):
			ber_tag = cls.ber_tag
		for key, value in cls.wrapped_clsattrs.items():
			setattr(WrappedType, key, value)
		value, rest = WrappedType.from_ber(data)
		return cls(value), rest

	@classmethod
	def to_ber(cls, obj):
		class WrappedType(cls.wrapped_type):
			ber_tag = cls.ber_tag
		for key, value in cls.wrapped_clsattrs.items():
			setattr(WrappedType, key, value)
		if not isinstance(obj, cls):
			raise TypeError()
		return WrappedType.to_ber(getattr(obj, cls.wrapped_attribute))

def retag(cls, tag):
	class Overwritten(cls):
		ber_tag = tag
	return Overwritten

class Enum(BERType):
	ber_tag = (0, False, 10)
	enum_type = None

	@classmethod
	def from_ber(cls, data):
		obj, rest = decode_ber(data)
		if obj.tag != cls.ber_tag:
			raise ValueError()
		value = decode_ber_integer(obj.content)
		return cls.enum_type(value), rest

	@classmethod
	def to_ber(cls, obj):
		if not isinstance(obj, cls.enum_type):
			raise TypeError()
		return encode_ber(BERObject(cls.ber_tag, encode_ber_integer(obj.value)))

def wrapenum(enumtype):
	class WrappedEnum(Enum):
		enum_type = enumtype
	return WrappedEnum



import traceback
import hashlib
import secrets
import socket
import ssl
from socketserver import BaseRequestHandler

from .ldap import *
from .asn1 import IncompleteBERError
from .directory import RootDSE, Subschema, BaseDirectory
from .schema import CORE_SUBSCHEMA
from .exceptions import *

def decode_msg(shallowmsg):
	try:
		return shallowmsg.decode()[0]
	except:
		traceback.print_exc()
		raise LDAPProtocolError()

def reject_critical_controls(controls=None, supported_oids=[]):
	for control in controls or []:
		if not control.criticality:
			continue
		if control.controlType not in supported_oids:
			raise LDAPUnavailableCriticalExtension()

class BaseLDAPRequestHandler(BaseRequestHandler):
	def setup(self):
		super().setup()
		self.keep_running = True

	def handle(self):
		self.on_connect()
		buf = b''
		while self.keep_running:
			try:
				shallowmsg, buf = ShallowLDAPMessage.from_ber(buf)
				for respmsg in self.handle_message(shallowmsg):
					self.request.sendall(LDAPMessage.to_ber(respmsg))
			except IncompleteBERError:
				chunk = self.request.recv(5)
				if not chunk:
					self.keep_running = False
					self.on_disconnect()
					self.request.close()
				else:
					buf += chunk
		self.on_disconnect()
		self.request.close()

	def handle_message(self, shallowmsg):
		'''Handle an LDAP request

		:param shallowmsg: Half-decoded LDAP message to handle
		:type shallowmsg: ShallowLDAPMessage
		:returns: Response messages
		:rtype: iterable of LDAPMessage objects
		'''
		msgtypes = {
			BindRequest: (self.handle_bind, BindResponse),
			UnbindRequest: (self.handle_unbind, None),
			SearchRequest: (self.handle_search, SearchResultDone),
			ModifyRequest: (self.handle_modify, ModifyResponse),
			AddRequest: (self.handle_add,  AddResponse),
			DelRequest: (self.handle_delete, DelResponse),
			ModifyDNRequest: (self.handle_modifydn, ModifyDNResponse),
			CompareRequest: (self.handle_compare, CompareResponse),
			AbandonRequest: (self.handle_abandon, None),
			ExtendedRequest: (self.handle_extended, ExtendedResponse),
		}
		handler, response_type = msgtypes.get(shallowmsg.protocolOpType, (None, None))
		try:
			if handler is None:
				raise LDAPProtocolError()
			try:
				msg = decode_msg(shallowmsg)
			except ValueError:
				self.on_recv_invalid(shallowmsg)
				raise LDAPProtocolError()
			self.on_recv(msg)
			for args in handler(msg.protocolOp, msg.controls):
				response, controls = args if isinstance(args, tuple) else (args, None)
				yield LDAPMessage(shallowmsg.messageID, response, controls)
		except LDAPError as e:
			if response_type is not None:
				respmsg = LDAPMessage(shallowmsg.messageID, response_type(e.code, diagnosticMessage=e.message))
				self.on_send(respmsg)
				yield respmsg
		except Exception as e:
			if response_type is not None:
				respmsg = LDAPMessage(shallowmsg.messageID, response_type(LDAPResultCode.other))
				self.on_send(respmsg)
				yield respmsg
			self.on_exception(e)

	def on_connect(self):
		pass

	def on_disconnect(self):
		pass

	def on_send(self, msg):
		pass

	def on_recv(self, msg):
		pass

	def on_recv_invalid(self, shallowmsg):
		pass

	def on_exception(self, e):
		traceback.print_exc()

	def handle_bind(self, op, controls=None):
		reject_critical_controls(controls)
		raise LDAPAuthMethodNotSupported()

	def handle_unbind(self, op, controls=None):
		reject_critical_controls(controls)
		self.keep_running = False

	def handle_search(self, op, controls=None):
		reject_critical_controls(controls)
		yield SearchResultDone(LDAPResultCode.success)

	def handle_modify(self, op, controls=None):
		reject_critical_controls(controls)
		raise LDAPInsufficientAccessRights()

	def handle_add(self, op, controls=None):
		reject_critical_controls(controls)
		raise LDAPInsufficientAccessRights()

	def handle_delete(self, op, controls=None):
		reject_critical_controls(controls)
		raise LDAPInsufficientAccessRights()

	def handle_modifydn(self, op, controls=None):
		reject_critical_controls(controls)
		raise LDAPInsufficientAccessRights()

	def handle_compare(self, op, controls=None):
		reject_critical_controls(controls)
		raise LDAPInsufficientAccessRights()

	def handle_abandon(self, op, controls=None):
		reject_critical_controls(controls)

	def handle_extended(self, op, controls=None):
		reject_critical_controls(controls)
		raise LDAPProtocolError()

class SimpleLDAPRequestHandler(BaseLDAPRequestHandler):
	'''
	.. py:attribute:: rootdse

		Special single-object :any:`BaseDirectory` that contains information
		about the server, such as supported extentions and SASL authentication
		mechansims. Attributes can be accessed in a dict-like fashion.
	'''

	subschema = CORE_SUBSCHEMA
	'''
	.. py:attribute:: subschema

		Special single-object :any:`BaseDirectory` that describes the schema.
		Per default the subschema includes standard syntaxes, standard matching
		rules and objectclasses/attributetypes for the rootdse and subschema.
		It does not include objectclasses/attributetypes for actual data
		(e.g. users and groups). See :any:`Subschema` for details.

		If `subschema` is not `None`, the subschemaSubentry attribute is
		automatically added to all results returned by :any:`do_search`.
	'''

	def setup(self):
		super().setup()
		self.rootdse = RootDSE()
		self.rootdse['objectClass'] = [b'top']
		self.rootdse['supportedSASLMechanisms'] = self.get_sasl_mechanisms
		self.rootdse['supportedExtension'] = self.get_extentions
		self.rootdse['supportedLDAPVersion'] = [b'3']
		self.bind_object = None
		self.bind_sasl_state = None

	def get_extentions(self):
		'''Get supported LDAP extentions

		:returns: OIDs of supported LDAP extentions
		:rtype: list of bytes objects

		Called whenever the root DSE attribute "supportedExtension" is queried.'''
		res = []
		if self.supports_starttls:
			res.append(EXT_STARTTLS_OID.encode())
		if self.supports_whoami:
			res.append(EXT_WHOAMI_OID.encode())
		if self.supports_password_modify:
			res.append(EXT_PASSWORD_MODIFY_OID.encode())
		return res

	def get_sasl_mechanisms(self):
		'''Get supported SASL mechanisms

		:returns: Names of supported SASL mechanisms
		:rtype: list of bytes objects

		SASL mechanism name are typically all-caps, like "EXTERNAL".
		
		Called whenever the root DSE attribute "supportedSASLMechanisms" is queried.'''
		res = []
		if self.supports_sasl_anonymous:
			res.append(b'ANONYMOUS')
		if self.supports_sasl_plain:
			res.append(b'PLAIN')
		if self.supports_sasl_external:
			res.append(b'EXTERNAL')
		return res

	def handle_bind(self, op, controls=None):
		reject_critical_controls(controls)
		if op.version != 3:
			raise LDAPProtocolError('Unsupported protocol version')
		auth = op.authentication
		# Resume ongoing SASL dialog
		if self.bind_sasl_state and isinstance(auth, SaslCredentials) \
				and auth.mechanism == self.bind_sasl_state[0]:
			mechanism, iterator = self.bind_sasl_state
			self.bind_sasl_state = None
			resp_code = LDAPResultCode.saslBindInProgress
			try:
				resp = iterator.send(auth.credentials)
				self.bind_sasl_state = (mechanism, iterator)
			except StopIteration as e:
				resp_code = LDAPResultCode.success
				self.bind_object, resp = e.value
			yield BindResponse(resp_code, serverSaslCreds=resp)
			return
		# If auth type or SASL method changed, abort SASL dialog
		self.bind_sasl_state = None
		if isinstance(auth, SimpleAuthentication):
			self.bind_object = self.do_bind_simple(op.name, auth.password)
			yield BindResponse(LDAPResultCode.success)
		elif isinstance(auth, SaslCredentials):
			ret = self.do_bind_sasl(auth.mechanism, auth.credentials)
			if isinstance(ret, tuple):
				self.bind_object, resp = ret
				yield BindResponse(LDAPResultCode.success, serverSaslCreds=resp)
				return
			iterator = iter(ret)
			resp_code = LDAPResultCode.saslBindInProgress
			try:
				resp = next(iterator)
				self.bind_sasl_state = (auth.mechanism, iterator)
			except StopIteration as e:
				resp_code = LDAPResultCode.success
				self.bind_object, resp = e.value
			yield BindResponse(resp_code, serverSaslCreds=resp)
		else:
			yield from super().handle_bind(op, controls)

	def do_bind_simple(self, dn='', password=b''):
		'''Do LDAP BIND with simple authentication

		:param dn: Distinguished name of object to be authenticated or empty
		:type dn: str
		:param password: Password, may be empty
		:type password: bytes

		:returns: Bind object
		:rtype: obj

		Delegates implementation to :any:`do_bind_simple_anonymous`,
		:any:`do_bind_simple_unauthenticated` or :any:`do_bind_simple_authenticated`
		according to `RFC 4513`_.'''
		if not dn and not password:
			return self.do_bind_simple_anonymous()
		if not password:
			return self.do_bind_simple_unauthenticated(dn)
		return self.do_bind_simple_authenticated(dn, password)

	def do_bind_simple_anonymous(self):
		'''Do LDAP BIND with simple anonymous authentication (`RFC 4513 5.1.1.`_)

		:raises LDAPError: if authentication failed

		:returns: Bind object on success
		:rtype: obj

		Calld by :any:`do_bind_simple`. Always returns None.'''
		return None

	def do_bind_simple_unauthenticated(self, dn):
		'''Do LDAP BIND with simple unauthenticated authentication (`RFC 4513 5.1.2.`_)

		:param dn: Distinguished name of the object to be authenticated
		:type dn: str

		:raises LDAPError: if authentication failed

		:returns: Bind object on success
		:rtype: obj

		Calld by :any:`do_bind_simple`. The default implementation always raises an
		:any:`LDAPInvalidCredentials` exception.'''
		raise LDAPInvalidCredentials()

	def do_bind_simple_authenticated(self, dn, password):
		'''Do LDAP BIND with simple name/password authentication (`RFC 4513 5.1.3.`_)

		:param dn: Distinguished name of the object to be authenticated
		:type dn: str
		:param password: Password for object
		:type dn: bytes

		:raises LDAPError: if authentication failed

		:returns: Bind object on success
		:rtype: obj

		Calld by :any:`do_bind_simple`. The default implementation always raises an
		`LDAPInvalidCredentials` exception.'''
		raise LDAPInvalidCredentials()

	def do_bind_sasl(self, mechanism, credentials=None, dn=None):
		'''Do LDAP BIND with SASL authentication (RFC 4513 and 4422)

		:param mechanism: Name of the selected SASL mechanism
		:type mechanism: str
		:param credentials: Initial client response
		:type credentials: bytes, optional
		:param dn: Distinguished name in LDAP BIND request, should be ignored for
		           SASL authentication
		:type dn: str, optional

		:returns: Bind object and final server challenge, only returns on success
		:rtype: Tuple (obj, bytes/None)

		The call only returns if authentication succeeded. In any other case,
		an appropriate :any:`LDAPError` is raised.

		Some SASL methods require additional challenge-response round trips. These
		can be achieved with the `yield` statement:

		    client_response = yield server_challenge

		Generally all server challenges and client responses can always be absent
		(indicated by None), empty (empty bytes object) or consist of any number
		of bytes. Whether a challenge or response may or must be absent or present
		is defined by the individual SASL mechanism.

		IANA list of SASL mechansims: https://www.iana.org/assignments/sasl-mechanisms/sasl-mechanisms.xhtml
		'''
		if not mechanism:
			# Request to abort current negotiation (RFC4513 5.2.1.2)
			raise LDAPAuthMethodNotSupported()
		if mechanism == 'ANONYMOUS' and self.supports_sasl_anonymous:
			if credentials is not None:
				credentials = credentials.decode()
			return self.do_bind_sasl_anonymous(trace_info=credentials), None
		if mechanism == 'PLAIN' and self.supports_sasl_plain:
			if credentials is None:
				raise LDAPProtocolError('Unsupported protocol version')
			authzid, authcid, password = credentials.split(b'\0', 2)
			return self.do_bind_sasl_plain(authcid.decode(), password.decode(), authzid.decode() or None), None
		if mechanism == 'EXTERNAL' and self.supports_sasl_external:
			if credentials is not None:
				credentials = credentials.decode()
			return self.do_bind_sasl_external(authzid=credentials), None
		raise LDAPAuthMethodNotSupported()

	supports_sasl_anonymous = False

	def do_bind_sasl_anonymous(self, trace_info=None):
		'''Do LDAP BIND with SASL "ANONYMOUS" mechanism (RFC 4505)

		:param trace_info: Trace information, either an email address or an
		                   opaque string that does not contain the '@' character
		:type trace_info: str, optional

		:raises LDAPError: if authentication failed

		:returns: Bind object on success
		:rtype: obj

		Calld by :any:`do_bind_sasl`. The default implementation raises an
		:any:`LDAPAuthMethodNotSupported` exception.'''
		raise LDAPAuthMethodNotSupported()

	supports_sasl_plain = False

	def do_bind_sasl_plain(self, identity, password, authzid=None):
		'''Do LDAP BIND with SASL "PLAIN" mechanism (RFC 4616)

		:param identity: Authentication identity (authcid)
		:type identity: str
		:param password: Password (passwd)
		:type password: str
		:param authzid: Authorization identity
		:type authzid: str, optional

		:raises LDAPError: if authentication failed

		:returns: Bind object on success
		:rtype: obj

		Calld by :any:`do_bind_sasl`. The default implementation raises an
		:any:`LDAPAuthMethodNotSupported` exception.'''
		raise LDAPAuthMethodNotSupported()

	supports_sasl_external = False

	def do_bind_sasl_external(self, authzid=None):
		'''Do LDAP BIND with SASL "EXTERNAL" mechanism (RFC 4422 and 4513)

		:param authzid: Authorization identity
		:type authzid: str, optional

		:raises LDAPError: if authentication failed

		:returns: Bind object on success
		:rtype: obj

		EXTERNAL is commonly used for TLS client certificate authentication or
		system user based authentication on UNIX sockets.

		Calld by :any:`do_bind_sasl`. The default implementation raises an
		:any:`LDAPAuthMethodNotSupported` exception.'''
		raise LDAPAuthMethodNotSupported()

	def handle_search(self, op, controls=None):
		for dn, attributes in self.do_search(op.baseObject, op.scope, op.filter):
			pattributes = [PartialAttribute(name, values) for name, values in attributes.items()]
			if 'subschemaSubentry' not in attributes and self.subschema is not None:
				pattributes.append(PartialAttribute('subschemaSubentry', [bytes(self.subschema.dn)]))
			yield SearchResultEntry(dn, pattributes)
		yield SearchResultDone(LDAPResultCode.success)

	def do_search(self, baseobj, scope, filter):
		'''Do LDAP SEARCH operation

		:param baseobj: Distinguished name of the LDAP entry relative to which the
		                search is to be performed
		:type baseobj: str
		:param scope: Search scope
		:type scope: SearchScope
		:param filter: Filter object
		:type filter: Filter

		:raises LDAPError: on error

		:returns: Iterable of dn, attributes tuples

		The default implementation returns matching objects from the root dse and
		the subschema.'''
		yield from self.rootdse.search(baseobj, scope, filter)
		yield from self.subschema.search(baseobj, scope, filter)

	def handle_unbind(self, op, controls=None):
		reject_critical_controls(controls)
		self.keep_running = False
		return []

	def handle_extended(self, op, controls=None):
		reject_critical_controls(controls)
		if op.requestName == EXT_STARTTLS_OID and self.supports_starttls:
			# StartTLS (RFC 4511)
			yield ExtendedResponse(LDAPResultCode.success, responseName=EXT_STARTTLS_OID)
			try:
				self.do_starttls()
			except Exception as e:
				traceback.print_exc()
				self.keep_running = False
		elif op.requestName == EXT_WHOAMI_OID and self.supports_whoami:
			# "Who am I?" Operation (RFC 4532)
			identity = (self.do_whoami() or '').encode()
			yield ExtendedResponse(LDAPResultCode.success, responseValue=identity)
		elif op.requestName == EXT_PASSWORD_MODIFY_OID and self.supports_password_modify:
			# Password Modify Extended Operation (RFC 3062)
			newpw = None
			if op.requestValue is None:
				newpw = self.do_passwd()
			else:
				decoded, _ = PasswdModifyRequestValue.from_ber(op.requestValue)
				newpw = self.do_passwd(decoded.userIdentity, decoded.oldPasswd, decoded.newPasswd)
			if newpw is None:
				yield ExtendedResponse(LDAPResultCode.success)
			else:
				encoded = PasswdModifyResponseValue.to_ber(PasswdModifyResponseValue(newpw))
				yield ExtendedResponse(LDAPResultCode.success, responseValue=encoded)
		else:
			yield from super().handle_extended(op, controls)

	#: :any:`ssl.SSLContext` for StartTLS
	ssl_context = None

	@property
	def supports_starttls(self):
		'''
		'''
		return self.ssl_context is not None and not isinstance(self.request, ssl.SSLSocket)

	def do_starttls(self):
		'''Do StartTLS extended operation (RFC 4511)

		Called by `handle_extended()` if :any:`supports_starttls` is True. The default
		implementation uses `ssl_context`.

		Note that the (success) response to the request is sent before this method
		is called. If a call to this method fails, the LDAP connection is
		immediately terminated.'''
		self.request = self.ssl_context.wrap_socket(self.request, server_side=True)

	#:
	supports_whoami = False

	def do_whoami(self):
		'''Do "Who am I?" extended operation (RFC 4532)

		:returns: Current authorization identity (authzid) or empty string for anonymous sessions
		:rtype: str

		Called by `handle_extended()` if `supports_whoami` is True. The default
		implementation always returns an empty string.'''
		return ''

	#:
	supports_password_modify = False

	def do_password_modify(self, user=None, old_password=None, new_password=None):
		'''Do password modify extended operation (RFC 3062)
		
		:param user: User the request relates to, may or may not be a
		             distinguished name. If absent, the request relates to the
		             user currently associated with the LDAP connection
		:type user: str, optional
		:param old_password: Current password of user
		:type old_password: bytes, optional
		:param new_password: Desired password for user
		:type new_password: bytes, optional
		
		Called by `handle_extended()` if :any:`supports_password_modify` is True. The
		default implementation always raises an :any:`LDAPUnwillingToPerform` error.'''
		raise LDAPUnwillingToPerform()



def encode_attribute(value):
	if isinstance(value, bool):
		value = b'TRUE' if value else b'FALSE'
	if isinstance(value, int):
		value = str(value)
	if isinstance(value, str):
		value = value.encode()
	if not isinstance(value, bytes):
		value = bytes(value)
	return value

class CaseInsensitiveKey(str):
	def __hash__(self):
		return hash(self.lower())

	def __eq__(self, value):
		return self.lower() == value.lower()

class CaseInsensitiveDict(dict):
	def __init__(self, *args, **kwargs):
		if len(args) == 1 and isinstance(args[0], dict):
			kwargs = {CaseInsensitiveKey(k): v for k, v in args[0].items()}
			args = []
		else:
			kwargs = {CaseInsensitiveKey(k): v for k, v in kwargs.items()}
			args = [(CaseInsensitiveKey(k), v) for k, v in args]
		super().__init__(*args, **kwargs)

	def __contains__(self, key):
		return super().__contains__(CaseInsensitiveKey(key))

	def __setitem__(self, key, value):
		super().__setitem__(CaseInsensitiveKey(key), value)

	def __getitem__(self, key):
		return super().__getitem__(CaseInsensitiveKey(key))

	def get(self, key, default=None):
		if key in self:
			return self[key]
		return default

class AttributeDict(CaseInsensitiveDict):
	def __getitem__(self, key):
		if key not in self:
			self[key] = []
		return super().__getitem__(key)




from .util import encode_attribute, AttributeDict
from .dn import DN
from .ldap import *

class BaseDirectory:
	'''Base class for LDAP directories'''

	def search(self, baseobj, scope, filter):
		'''Perform search

		:param baseobj: Distinguished name of the LDAP entry relative to which the search is
		                to be performed
		:type baseobj: str
		:param scope: Search scope
		:type scope: SearchScope
		:param filter: Filter object
		:type filter: Filter

		:returns: Iterable of dn, attributes tuples'''
		return []

class FilterMixin:
	'''Mixin for :any:`BaseDirectory` that implements :any:`BaseDirectory.search` by calling appropirate `filter_*` methods'''
	def search(self, baseobj, scope, filter):
		dn_res = self.filter_dn(baseobj, scope)
		filter_res = self.search_filter(filter)
		return self.search_fetch(self.filter_and(dn_res, filter_res))

	def search_fetch(self, result):
		'''
		'''
		return []

	def search_filter(self, expr):
		'''
		'''
		if isinstance(expr, FilterAnd):
			return self.filter_and(*[self.search_filter(subexpr) for subexpr in expr.filters])
		elif isinstance(expr, FilterOr):
			return self.filter_or(*[self.search_filter(subexpr) for subexpr in expr.filters])
		elif isinstance(expr, FilterNot):
			return self.filter_not(self.search_filter(expr.filter))
		elif isinstance(expr, FilterEqual):
			return self.filter_equal(expr.attribute.lower(), expr.value)
		elif isinstance(expr, FilterPresent):
			return self.filter_present(expr.attribute.lower())
		else:
			return False

	def filter_present(self, attribute):
		'''
		'''
		return False

	def filter_equal(self, attribute, value):
		'''
		'''
		return False

	def filter_and(self, *subresults):
		'''
		'''
		filtered = []
		for subres in subresults:
			if subres is True:
				continue
			if subres is False:
				return False
			filtered.append(subres)
		if not filtered:
			return True
		if len(filtered) == 1:
			return filtered[0]
		return self._filter_and(*filtered)

	def _filter_and(self, *subresults):
		'''
		'''
		return False

	def filter_or(self, *subresults):
		'''
		'''
		filtered = []
		for subres in subresults:
			if subres is True:
				return True
			if subres is False:
				continue
			filtered.append(subres)
		if not filtered:
			return False
		if len(filtered) == 1:
			return filtered[0]
		return self._filter_or(*filtered)

	def _filter_or(self, *subresults):
		'''
		'''
		return False

	def filter_not(self, subresult):
		'''
		'''
		if subresult is True:
			return False
		if subresult is False:
			return True
		return self._filter_not(subresult)

	def _filter_not(self, subresult):
		'''
		'''
		return False

	def filter_dn(self, baseobj, scope):
		'''
		'''
		return False

class SimpleFilterMixin(FilterMixin):
	def filter_present(self, attribute):
		if attribute in ['objectclass', 'structuralobjectclass', 'subschemasubentry']:
			return True
		return FilterPresent(attribute)

	def filter_equal(self, attribute, value):
		if attribute == 'objectclass':
			return value.lower() in [s.lower() for s in self.objectclasses]
		elif attribute == 'structuralobjectclass':
			return value.lower() == self.structuralobjectclass.lower()
		return FilterEqual(attribute, value)

	def _filter_and(self, *subresults):
		return FilterAnd(subresults)

	def _filter_or(self, *subresults):
		return FilterOr(subresults)

	def _filter_not(self, subresult):
		return FilterNot(subresult)

	def filter_dn(self, base, scope):
		base = DN(base)
		if scope == SearchScope.baseObject:
			if base[1:] != self.dn_base or len(base[0]) != 1 or base[0][0].attribute != self.rdn_attr:
				return False
			return self.filter_equal(self.rdn_attr, base[0][0].value)
		elif scope == SearchScope.singleLevel:
			return base == self.dn_base
		elif scope == SearchScope.wholeSubtree:
			if self.dn_base.in_subtree_of(base):
				return True
			if base[1:] != self.dn_base or len(base[0]) != 1 or base[0][0].attribute != self.rdn_attr:
				return False
			return self.filter_equal(self.rdn_attr, base[0][0].value)
		else:
			return False

class RootDSE(BaseDirectory, AttributeDict):
	def search(self, baseobj, scope, filter):
		if baseobj or scope != SearchScope.baseObject:
			return []
		if not isinstance(filter, FilterPresent) or filter.attribute.lower() != 'objectclass':
			return []
		attrs = {}
		for name, values in self.items():
			if callable(values):
				values = values()
			if not isinstance(values, list):
				values = [values]
			if values:
				attrs[name] = [encode_attribute(value) for value in values]
		return [('', attrs)]

class Subschema(BaseDirectory, AttributeDict):
	def __init__(self, *args,
	             dn='cn=Subschema',
	             structuralobjectclass='subentry',
	             objectclass=('top', 'subschema', 'extensibleObject'),
	             subtreespecification='{ }',
	             ldapsyntaxes=tuple(),
	             matchingrules=tuple(),
	             objectclasses=tuple(),
	             attributetypes=tuple(),
	             **kwargs):
		super().__init__(*args, **kwargs)
		self.dn = DN(dn)
		self['structuralObjectClass'] = [structuralobjectclass]
		self['objectClass'] = list(objectclass)
		self['subtreeSpecification'] = [subtreespecification]
		self['ldapSyntaxes'] = list(ldapsyntaxes)
		self['matchingRules'] = list(matchingrules)
		self['objectClasses'] = list(objectclasses)
		self['attributeTypes'] = list(attributetypes)

	def search(self, baseobj, scope, filter):
		if DN(baseobj) != self.dn or scope != SearchScope.baseObject:
			return []
		if not isinstance(filter, FilterEqual):
			return []
		if filter.attribute.lower() != 'objectclass' or filter.value.lower() != b'subschema':
			return []
		return [(str(self.dn), {key: [encode_attribute(value) for value in values] for key, values in self.items()})]

def eval_ldap_filter(obj, expr):
	'''Return whether LDAP filter expression matches attribute dictionary'''
	if expr is True:
		return True
	elif expr is False:
		return False
	elif isinstance(expr, FilterAnd):
		for subexpr in expr.filters:
			if not eval_ldap_filter(obj, subexpr):
				return False
		return True
	elif isinstance(expr, FilterOr):
		for subexpr in expr.filters:
			if eval_ldap_filter(obj, subexpr):
				return True
		return False
	elif isinstance(expr, FilterNot):
		return not eval_ldap_filter(obj, expr.filter)
	elif isinstance(expr, FilterEqual):
		return expr.value in obj.get(expr.attribute, [])
	elif isinstance(expr, FilterPresent):
		return bool(obj.get(expr.attribute, []))
	else:
		return False

class StaticDirectory(BaseDirectory):
	def __init__(self):
		self.objects = {} # dn -> attribute dict

	def add(self, dn, attributes):
		tmp = AttributeDict()
		for key, values in attributes.items():
			tmp[key] = [encode_attribute(value) for value in values]
		self.objects[DN(dn)] = tmp

	def search(self, baseobj, scope, filter):
		baseobj = DN(baseobj)
		for dn, attributes in self.objects.items():
			if scope == SearchScope.baseObject:
				if baseobj != dn:
					continue
			elif scope == SearchScope.singleLevel:
				if not dn.is_direct_child_of(baseobj):
					continue
			elif scope == SearchScope.wholeSubtree:
				if not dn.in_subtree_of(baseobj):
					continue
			else:
				continue
			if not eval_ldap_filter(attributes, filter):
				continue
			yield str(dn), attributes



from string import hexdigits as HEXDIGITS

from .util import encode_attribute

DN_ESCAPED = ('"', '+', ',', ';', '<', '>')
DN_SPECIAL = DN_ESCAPED + (' ', '#', '=')

def parse_assertion(expr, case_ignore_attrs=None):
	case_ignore_attrs = case_ignore_attrs or []
	hexdigit = None
	escaped = False
	tokens = []
	token = b''
	for c in expr:
		if hexdigit is not None:
			if c not in HEXDIGITS:
				raise ValueError('Invalid hexpair: \\%c%c'%(hexdigit, c))
			token += bytes.fromhex('%c%c'%(hexdigit, c))
			hexdigit = None
		elif escaped:
			escaped = False
			if c in DN_SPECIAL or c == '\\':
				token += c.encode()
			elif c in HEXDIGITS:
				hexdigit = c
			else:
				raise ValueError('Invalid escape: \\%c'%c)
		elif c == '\\':
			escaped = True
		elif c == '=':
			tokens.append(token)
			token = b''
		else:
			token += c.encode()
	tokens.append(token)
	if len(tokens) != 2:
		raise ValueError('Invalid assertion in RDN: "%s"'%expr)
	name = tokens[0].decode().lower()
	value = tokens[1]
	if not name or not value:
		raise ValueError('Invalid assertion in RDN: "%s"'%expr)
	# TODO: handle hex strings
	if name in case_ignore_attrs:
		value = value.lower()
	return (name, value)

def parse_rdn(rdn, case_ignore_attrs=None):
	escaped = False
	assertions = []
	token = ''
	for c in rdn:
		if escaped:
			escaped = False
			token += c
		elif c == '+':
			assertions.append(parse_assertion(token, case_ignore_attrs=case_ignore_attrs))
			token = ''
		else:
			if c == '\\':
				escaped = True
			token += c
	assertions.append(parse_assertion(token, case_ignore_attrs=case_ignore_attrs))
	if not assertions:
		raise ValueError('Invalid RDN "%s"'%rdn)
	return tuple(sorted(assertions))

def parse_dn(dn, case_ignore_attrs=None):
	if not dn:
		return tuple()
	escaped = False
	rdns = []
	rdn = ''
	for c in dn:
		if escaped:
			escaped = False
			rdn += c
		elif c == ',':
			rdns.append(parse_rdn(rdn, case_ignore_attrs=case_ignore_attrs))
			rdn = ''
		else:
			if c == '\\':
				escaped = True
			rdn += c
	rdns.append(parse_rdn(rdn, case_ignore_attrs=case_ignore_attrs))
	return tuple(rdns)

# >>> parse_dn('OU=Sales+CN=J.  Smith,DC=example,DC=net', case_ignore_attrs=['cn', 'ou', 'dc'])
# ((('cn', b'j.  smith'), ('ou', b'sales')), (('dc', b'example'),), (('dc', b'net'),))

def escape_dn_value(value):
	if isinstance(value, int):
		value = str(value)
	if isinstance(value, str):
		value = value.encode()
	res = ''
	for c in value:
		c = bytes((c,))
		try:
			s = c.decode()
		except UnicodeDecodeError:
			s = '\\'+c.hex()
		if s in DN_SPECIAL:
			s = '\\'+s
		res += s
	return res

def build_assertion(assertion):
	name, value = assertion
	return '%s=%s'%(escape_dn_value(name.encode()), escape_dn_value(value))

def build_rdn(assertions):
	return '+'.join(map(build_assertion, assertions))

def build_dn(rdns):
	return ','.join(map(build_rdn, rdns))

class DN(tuple):
	def __new__(cls, *args):
		if len(args) == 1 and isinstance(args[0], str):
			return super().__new__(cls, [RDN(*rdn) for rdn in parse_dn(args[0])])
		elif len(args) == 1 and isinstance(args[0], DN):
			return args[0]
		else:
			return super().__new__(cls, [RDN(rdn) for rdn in args])

	def __repr__(self):
		return '<DN(%s)>'%repr(str(self))

	def __str__(self):
		return build_dn(self)

	def __bytes__(self):
		return str(self).encode()

	def __add__(self, value):
		if isinstance(value, DN):
			return DN(*(tuple(self) + tuple(value)))
		else:
			raise ValueError()

	def __getitem__(self, key):
		if isinstance(key, slice):
			return DN(*super().__getitem__(key))
		return super().__getitem__(key)

	def strip_common_suffix(self, value):
		value = DN(value)
		minlen = min(len(self), len(value))
		for i in range(minlen):
			if self[-1 - i] != value[-1 - i]:
				return self[:-i or None], value[:-i or None]
		return self[:-minlen], value[:-minlen]

	def is_direct_child_of(self, base):
		rchild, rbase = self.strip_common_suffix(DN(base))
		return not rbase and len(rchild) == 1

	def in_subtree_of(self, base):
		rchild, rbase = self.strip_common_suffix(DN(base))
		return not rbase

class RDN(tuple):
	def __new__(cls, *args, **kwargs):
		if not kwargs and len(args) == 1 and isinstance(args[0], str):
			return super().__new__(cls, [RDNAssertion(ava) for ava in parse_rdn(args[0])])
		elif not kwargs and len(args) == 1 and isinstance(args[0], cls):
			return args[0]
		else:
			assertions = []
			for key, value in args + tuple(kwargs.items()):
				assertion = RDNAssertion(key, value)
				if assertion not in assertions:
					assertions.append(assertion)
			assertions.sort()
			return super().__new__(cls, assertions)

	def __repr__(self):
		return '<RDN(%s)>'%repr(str(self))

	def __str__(self):
		return build_rdn(self)

class RDNAssertion(tuple):
	def __new__(cls, *args):
		if len(args) == 1 and isinstance(args[0], RDNAssertion):
			return args[0]
		elif len(args) == 1 and isinstance(args[0], str):
			return super().__new__(cls, parse_assertion(args[0]))
		else:
			key, value = args
			value = encode_attribute(value)
			if not isinstance(key, str):
				raise ValueError('Key in RDN assertion "%s=%s" has invalid type'%(repr(key), repr(value)))
			if not isinstance(value, bytes):
				raise ValueError('Value in RDN assertion "%s=%s" has invalid type'%(key, repr(value)))
			return super().__new__(cls, (key.lower(), value))

	def __repr__(self):
		return '<RDNAssertion(%s)>'%repr(str(self))

	def __str__(self):
		return build_assertion(self)

	@property
	def attribute(self):
		return self[0]

	@property
	def value(self):
		return self[1]



from . import directory

CORE_SYNTAXES = (
	# RFC 4517
	"( 1.3.6.1.4.1.1466.115.121.1.3 DESC 'Attribute Type Description' )",
	"( 1.3.6.1.4.1.1466.115.121.1.6 DESC 'Bit String' )",
	"( 1.3.6.1.4.1.1466.115.121.1.7 DESC 'Boolean' )",
	"( 1.3.6.1.4.1.1466.115.121.1.11 DESC 'Country String' )",
	"( 1.3.6.1.4.1.1466.115.121.1.14 DESC 'Delivery Method' )",
	"( 1.3.6.1.4.1.1466.115.121.1.15 DESC 'Directory String' )",
	"( 1.3.6.1.4.1.1466.115.121.1.16 DESC 'DIT Content Rule Description' )",
	"( 1.3.6.1.4.1.1466.115.121.1.17 DESC 'DIT Structure Rule Description' )",
	"( 1.3.6.1.4.1.1466.115.121.1.12 DESC 'DN' )",
	"( 1.3.6.1.4.1.1466.115.121.1.21 DESC 'Enhanced Guide' )",
	"( 1.3.6.1.4.1.1466.115.121.1.22 DESC 'Facsimile Telephone Number')",
	"( 1.3.6.1.4.1.1466.115.121.1.23 DESC 'Fax' )",
	"( 1.3.6.1.4.1.1466.115.121.1.24 DESC 'Generalized Time' )",
	"( 1.3.6.1.4.1.1466.115.121.1.25 DESC 'Guide' )",
	"( 1.3.6.1.4.1.1466.115.121.1.26 DESC 'IA5 String' )",
	"( 1.3.6.1.4.1.1466.115.121.1.27 DESC 'INTEGER' )",
	"( 1.3.6.1.4.1.1466.115.121.1.28 DESC 'JPEG' )",
	"( 1.3.6.1.4.1.1466.115.121.1.54 DESC 'LDAP Syntax Description' )",
	"( 1.3.6.1.4.1.1466.115.121.1.30 DESC 'Matching Rule Description' )",
	"( 1.3.6.1.4.1.1466.115.121.1.31 DESC 'Matching Rule Use Description' )",
	"( 1.3.6.1.4.1.1466.115.121.1.34 DESC 'Name And Optional UID' )",
	"( 1.3.6.1.4.1.1466.115.121.1.35 DESC 'Name Form Description' )",
	"( 1.3.6.1.4.1.1466.115.121.1.36 DESC 'Numeric String' )",
	"( 1.3.6.1.4.1.1466.115.121.1.37 DESC 'Object Class Description' )",
	"( 1.3.6.1.4.1.1466.115.121.1.40 DESC 'Octet String' )",
	"( 1.3.6.1.4.1.1466.115.121.1.38 DESC 'OID' )",
	"( 1.3.6.1.4.1.1466.115.121.1.39 DESC 'Other Mailbox' )",
	"( 1.3.6.1.4.1.1466.115.121.1.41 DESC 'Postal Address' )",
	"( 1.3.6.1.4.1.1466.115.121.1.44 DESC 'Printable String' )",
	"( 1.3.6.1.4.1.1466.115.121.1.58 DESC 'Substring Assertion' )",
	"( 1.3.6.1.4.1.1466.115.121.1.50 DESC 'Telephone Number' )",
	"( 1.3.6.1.4.1.1466.115.121.1.51 DESC 'Teletex Terminal Identifier' )",
	"( 1.3.6.1.4.1.1466.115.121.1.52 DESC 'Telex Number' )",
	"( 1.3.6.1.4.1.1466.115.121.1.53 DESC 'UTC Time' )",
	# RFC 3672 (Subentries in LDAP)
	"( 1.3.6.1.4.1.1466.115.121.1.45 DESC 'SubtreeSpecification' )",
)

CORE_MATCHING_RULES = (
	# RFC 4517
	"( 2.5.13.16 NAME 'bitStringMatch' SYNTAX 1.3.6.1.4.1.1466.115.121.1.6 )",
	"( 2.5.13.13 NAME 'booleanMatch' SYNTAX 1.3.6.1.4.1.1466.115.121.1.7 )",
	"( 1.3.6.1.4.1.1466.109.114.1 NAME 'caseExactIA5Match' SYNTAX 1.3.6.1.4.1.1466.115.121.1.26 )",
	"( 2.5.13.5 NAME 'caseExactMatch' SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 )",
	"( 2.5.13.6 NAME 'caseExactOrderingMatch' SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 )",
	"( 2.5.13.7 NAME 'caseExactSubstringsMatch' SYNTAX 1.3.6.1.4.1.1466.115.121.1.58 )",
	"( 1.3.6.1.4.1.1466.109.114.2 NAME 'caseIgnoreIA5Match' SYNTAX 1.3.6.1.4.1.1466.115.121.1.26 )",
	"( 1.3.6.1.4.1.1466.109.114.3 NAME 'caseIgnoreIA5SubstringsMatch' SYNTAX 1.3.6.1.4.1.1466.115.121.1.58 )",
	"( 2.5.13.11 NAME 'caseIgnoreListMatch' SYNTAX 1.3.6.1.4.1.1466.115.121.1.41 )",
	"( 2.5.13.12 NAME 'caseIgnoreListSubstringsMatch' SYNTAX 1.3.6.1.4.1.1466.115.121.1.58 )",
	"( 2.5.13.2 NAME 'caseIgnoreMatch' SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 )",
	"( 2.5.13.3 NAME 'caseIgnoreOrderingMatch' SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 )",
	"( 2.5.13.4 NAME 'caseIgnoreSubstringsMatch' SYNTAX 1.3.6.1.4.1.1466.115.121.1.58 )",
	"( 2.5.13.31 NAME 'directoryStringFirstComponentMatch' SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 )",
	"( 2.5.13.1 NAME 'distinguishedNameMatch' SYNTAX 1.3.6.1.4.1.1466.115.121.1.12 )",
	"( 2.5.13.27 NAME 'generalizedTimeMatch' SYNTAX 1.3.6.1.4.1.1466.115.121.1.24 )",
	"( 2.5.13.28 NAME 'generalizedTimeOrderingMatch' SYNTAX 1.3.6.1.4.1.1466.115.121.1.24 )",
	"( 2.5.13.29 NAME 'integerFirstComponentMatch' SYNTAX 1.3.6.1.4.1.1466.115.121.1.27 )",
	"( 2.5.13.14 NAME 'integerMatch' SYNTAX 1.3.6.1.4.1.1466.115.121.1.27 )",
	"( 2.5.13.15 NAME 'integerOrderingMatch' SYNTAX 1.3.6.1.4.1.1466.115.121.1.27 )",
	"( 2.5.13.33 NAME 'keywordMatch' SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 )",
	"( 2.5.13.8 NAME 'numericStringMatch' SYNTAX 1.3.6.1.4.1.1466.115.121.1.36 )",
	"( 2.5.13.9 NAME 'numericStringOrderingMatch' SYNTAX 1.3.6.1.4.1.1466.115.121.1.36 )",
	"( 2.5.13.10 NAME 'numericStringSubstringsMatch' SYNTAX 1.3.6.1.4.1.1466.115.121.1.58 )",
	"( 2.5.13.30 NAME 'objectIdentifierFirstComponentMatch' SYNTAX 1.3.6.1.4.1.1466.115.121.1.38 )",
	"( 2.5.13.0 NAME 'objectIdentifierMatch' SYNTAX 1.3.6.1.4.1.1466.115.121.1.38 )",
	"( 2.5.13.17 NAME 'octetStringMatch' SYNTAX 1.3.6.1.4.1.1466.115.121.1.40 )",
	"( 2.5.13.18 NAME 'octetStringOrderingMatch' SYNTAX 1.3.6.1.4.1.1466.115.121.1.40 )",
	"( 2.5.13.20 NAME 'telephoneNumberMatch' SYNTAX 1.3.6.1.4.1.1466.115.121.1.50 )",
	"( 2.5.13.21 NAME 'telephoneNumberSubstringsMatch' SYNTAX 1.3.6.1.4.1.1466.115.121.1.58 )",
	"( 2.5.13.23 NAME 'uniqueMemberMatch' SYNTAX 1.3.6.1.4.1.1466.115.121.1.34 )",
	"( 2.5.13.32 NAME 'wordMatch' SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 )",
)

# Self-contained object class and attribute type Definitions that describe
# everthing need for the root DSE and subschema entry

CORE_OBJECT_CLASSES = (
	# RFC 4512
	"( 2.5.6.0 NAME 'top' ABSTRACT MUST objectClass )",
	"( 2.5.6.1 NAME 'alias' SUP top STRUCTURAL MUST aliasedObjectName )",
	"( 2.5.20.1 NAME 'subschema' AUXILIARY MAY ( dITStructureRules $ nameForms $ ditContentRules $ objectClasses $ attributeTypes $ matchingRules $ matchingRuleUse ) )",
	"( 1.3.6.1.4.1.1466.101.120.111 NAME 'extensibleObject' SUP top AUXILIARY )",
	# RFC 3672 (Subentries in LDAP)
	"( 2.5.17.0 NAME 'subentry' SUP top STRUCTURAL MUST ( cn $ subtreeSpecification ) )",
)

CORE_ATTRIBUTE_TYPES = (
	# RFC 4512
	"( 2.5.4.1 NAME 'aliasedObjectName' EQUALITY distinguishedNameMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.12 SINGLE-VALUE )",
	"( 2.5.4.0 NAME 'objectClass' EQUALITY objectIdentifierMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.38 )",
	"( 2.5.18.3 NAME 'creatorsName' EQUALITY distinguishedNameMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.12 SINGLE-VALUE NO-USER-MODIFICATION USAGE directoryOperation )",
	"( 2.5.18.1 NAME 'createTimestamp' EQUALITY generalizedTimeMatch ORDERING generalizedTimeOrderingMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.24 SINGLE-VALUE NO-USER-MODIFICATION USAGE directoryOperation )",
	"( 2.5.18.4 NAME 'modifiersName' EQUALITY distinguishedNameMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.12 SINGLE-VALUE NO-USER-MODIFICATION USAGE directoryOperation )",
	"( 2.5.18.2 NAME 'modifyTimestamp' EQUALITY generalizedTimeMatch ORDERING generalizedTimeOrderingMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.24 SINGLE-VALUE NO-USER-MODIFICATION USAGE directoryOperation )",
	"( 2.5.21.9 NAME 'structuralObjectClass' EQUALITY objectIdentifierMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.38 SINGLE-VALUE NO-USER-MODIFICATION USAGE directoryOperation )",
	"( 2.5.21.10 NAME 'governingStructureRule' EQUALITY integerMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.27 SINGLE-VALUE NO-USER-MODIFICATION USAGE directoryOperation )",
	"( 2.5.18.10 NAME 'subschemaSubentry' EQUALITY distinguishedNameMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.12 SINGLE-VALUE NO-USER-MODIFICATION USAGE directoryOperation )",
	"( 2.5.21.6 NAME 'objectClasses' EQUALITY objectIdentifierFirstComponentMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.37 USAGE directoryOperation )",
	"( 2.5.21.5 NAME 'attributeTypes' EQUALITY objectIdentifierFirstComponentMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.3 USAGE directoryOperation )",
	"( 2.5.21.4 NAME 'matchingRules' EQUALITY objectIdentifierFirstComponentMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.30 USAGE directoryOperation )",
	"( 2.5.21.8 NAME 'matchingRuleUse' EQUALITY objectIdentifierFirstComponentMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.31 USAGE directoryOperation )",
	"( 1.3.6.1.4.1.1466.101.120.16 NAME 'ldapSyntaxes' EQUALITY objectIdentifierFirstComponentMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.54 USAGE directoryOperation )",
	"( 2.5.21.2 NAME 'dITContentRules' EQUALITY objectIdentifierFirstComponentMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.16 USAGE directoryOperation )",
	"( 2.5.21.1 NAME 'dITStructureRules' EQUALITY integerFirstComponentMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.17 USAGE directoryOperation )",
	"( 2.5.21.7 NAME 'nameForms' EQUALITY objectIdentifierFirstComponentMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.35 USAGE directoryOperation )",
	"( 1.3.6.1.4.1.1466.101.120.6 NAME 'altServer' SYNTAX 1.3.6.1.4.1.1466.115.121.1.26 USAGE dSAOperation )",
	"( 1.3.6.1.4.1.1466.101.120.5 NAME 'namingContexts' SYNTAX 1.3.6.1.4.1.1466.115.121.1.12 USAGE dSAOperation )",
	"( 1.3.6.1.4.1.1466.101.120.13 NAME 'supportedControl' SYNTAX 1.3.6.1.4.1.1466.115.121.1.38 USAGE dSAOperation )",
	"( 1.3.6.1.4.1.1466.101.120.7 NAME 'supportedExtension' SYNTAX 1.3.6.1.4.1.1466.115.121.1.38 USAGE dSAOperation )",
	"( 1.3.6.1.4.1.4203.1.3.5 NAME 'supportedFeatures' EQUALITY objectIdentifierMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.38 USAGE dSAOperation )",
	"( 1.3.6.1.4.1.1466.101.120.15 NAME 'supportedLDAPVersion' SYNTAX 1.3.6.1.4.1.1466.115.121.1.27 USAGE dSAOperation )",
	"( 1.3.6.1.4.1.1466.101.120.14 NAME 'supportedSASLMechanisms' SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 USAGE dSAOperation )",
	# RFC 3672 (Subentries in LDAP)
	"( 2.5.18.5 NAME 'administrativeRole' EQUALITY objectIdentifierMatch USAGE directoryOperation SYNTAX 1.3.6.1.4.1.1466.115.121.1.38 )",
	"( 2.5.18.6 NAME 'subtreeSpecification' SINGLE-VALUE USAGE directoryOperation SYNTAX 1.3.6.1.4.1.1466.115.121.1.45 )",
	# RFC 4519 (Schema for User Applications)
	"( 2.5.4.41 NAME 'name' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 )",
	"( 2.5.4.3 NAME 'cn' SUP name )",
)

CORE_SUBSCHEMA = directory.Subschema(ldapsyntaxes=CORE_SYNTAXES,
                                     matchingrules=CORE_MATCHING_RULES,
                                     attributetypes=CORE_ATTRIBUTE_TYPES,
                                     objectclasses=CORE_OBJECT_CLASSES)

# RFC 4519: Schema for User Applications

RFC4519_ATTRIBUTE_TYPES = CORE_ATTRIBUTE_TYPES + (
	"( 2.5.4.15 NAME 'businessCategory' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 )",
	"( 2.5.4.6 NAME 'c' SUP name SYNTAX 1.3.6.1.4.1.1466.115.121.1.11 SINGLE-VALUE )",
	# Already included in CORE_ATTRIBUTE_TYPES
	#"( 2.5.4.3 NAME 'cn' SUP name )",
	"( 0.9.2342.19200300.100.1.25 NAME 'dc' EQUALITY caseIgnoreIA5Match SUBSTR caseIgnoreIA5SubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.26 SINGLE-VALUE )",
	"( 2.5.4.13 NAME 'description' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 )",
	"( 2.5.4.27 NAME 'destinationIndicator' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.44 )",
	"( 2.5.4.49 NAME 'distinguishedName' EQUALITY distinguishedNameMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.12 )",
	"( 2.5.4.46 NAME 'dnQualifier' EQUALITY caseIgnoreMatch ORDERING caseIgnoreOrderingMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.44 )",
	"( 2.5.4.47 NAME 'enhancedSearchGuide' SYNTAX 1.3.6.1.4.1.1466.115.121.1.21 )",
	"( 2.5.4.23 NAME 'facsimileTelephoneNumber' SYNTAX 1.3.6.1.4.1.1466.115.121.1.22 )",
	"( 2.5.4.44 NAME 'generationQualifier' SUP name )",
	"( 2.5.4.42 NAME 'givenName' SUP name )",
	"( 2.5.4.51 NAME 'houseIdentifier' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 )",
	"( 2.5.4.43 NAME 'initials' SUP name )",
	"( 2.5.4.25 NAME 'internationalISDNNumber' EQUALITY numericStringMatch SUBSTR numericStringSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.36 )",
	"( 2.5.4.7 NAME 'l' SUP name )",
	"( 2.5.4.31 NAME 'member' SUP distinguishedName )",
	# Already included in CORE_ATTRIBUTE_TYPES
	#"( 2.5.4.41 NAME 'name' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 )",
	"( 2.5.4.10 NAME 'o' SUP name )",
	"( 2.5.4.11 NAME 'ou' SUP name )",
	"( 2.5.4.32 NAME 'owner' SUP distinguishedName )",
	"( 2.5.4.19 NAME 'physicalDeliveryOfficeName' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 )",
	"( 2.5.4.16 NAME 'postalAddress' EQUALITY caseIgnoreListMatch SUBSTR caseIgnoreListSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.41 )",
	"( 2.5.4.17 NAME 'postalCode' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 )",
	"( 2.5.4.18 NAME 'postOfficeBox' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 )",
	"( 2.5.4.28 NAME 'preferredDeliveryMethod' SYNTAX 1.3.6.1.4.1.1466.115.121.1.14 SINGLE-VALUE )",
	"( 2.5.4.26 NAME 'registeredAddress' SUP postalAddress SYNTAX 1.3.6.1.4.1.1466.115.121.1.41 )",
	"( 2.5.4.33 NAME 'roleOccupant' SUP distinguishedName )",
	"( 2.5.4.14 NAME 'searchGuide' SYNTAX 1.3.6.1.4.1.1466.115.121.1.25 )",
	"( 2.5.4.34 NAME 'seeAlso' SUP distinguishedName )",
	"( 2.5.4.5 NAME 'serialNumber' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.44 )",
	"( 2.5.4.4 NAME 'sn' SUP name )",
	"( 2.5.4.8 NAME 'st' SUP name )",
	"( 2.5.4.9 NAME 'street' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 )",
	"( 2.5.4.20 NAME 'telephoneNumber' EQUALITY telephoneNumberMatch SUBSTR telephoneNumberSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.50 )",
	"( 2.5.4.22 NAME 'teletexTerminalIdentifier' SYNTAX 1.3.6.1.4.1.1466.115.121.1.51 )",
	"( 2.5.4.21 NAME 'telexNumber' SYNTAX 1.3.6.1.4.1.1466.115.121.1.52 )",
	"( 2.5.4.12 NAME 'title' SUP name )",
	"( 0.9.2342.19200300.100.1.1 NAME 'uid' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 )",
	"( 2.5.4.50 NAME 'uniqueMember' EQUALITY uniqueMemberMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.34 )",
	"( 2.5.4.35 NAME 'userPassword' EQUALITY octetStringMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.40 )",
	"( 2.5.4.24 NAME 'x121Address' EQUALITY numericStringMatch SUBSTR numericStringSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.36 )",
	"( 2.5.4.45 NAME 'x500UniqueIdentifier' EQUALITY bitStringMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.6 )",
)

RFC4519_OBJECT_CLASSES = CORE_OBJECT_CLASSES + (
	"( 2.5.6.11 NAME 'applicationProcess' SUP top STRUCTURAL MUST cn MAY ( seeAlso $ ou $ l $ description ) )",
	"( 2.5.6.2 NAME 'country' SUP top STRUCTURAL MUST c MAY ( searchGuide $ description ) )",
	"( 1.3.6.1.4.1.1466.344 NAME 'dcObject' SUP top AUXILIARY MUST dc )",
	"( 2.5.6.14 NAME 'device' SUP top STRUCTURAL MUST cn MAY ( serialNumber $ seeAlso $ owner $ ou $ o $ l $ description ) )",
	"( 2.5.6.9 NAME 'groupOfNames' SUP top STRUCTURAL MUST ( member $ cn ) MAY ( businessCategory $ seeAlso $ owner $ ou $ o $ description ) )",
	"( 2.5.6.17 NAME 'groupOfUniqueNames' SUP top STRUCTURAL MUST ( uniqueMember $ cn ) MAY ( businessCategory $ seeAlso $ owner $ ou $ o $ description ) )",
	"( 2.5.6.3 NAME 'locality' SUP top STRUCTURAL MAY ( street $ seeAlso $ searchGuide $ st $ l $ description ) )",
	"( 2.5.6.4 NAME 'organization' SUP top STRUCTURAL MUST o MAY ( userPassword $ searchGuide $ seeAlso $ businessCategory $ x121Address $ registeredAddress $ destinationIndicator $ preferredDeliveryMethod $ telexNumber $ teletexTerminalIdentifier $ telephoneNumber $ internationalISDNNumber $ facsimileTelephoneNumber $ street $ postOfficeBox $ postalCode $ postalAddress $ physicalDeliveryOfficeName $ st $ l $ description ) )",
	"( 2.5.6.7 NAME 'organizationalPerson' SUP person STRUCTURAL MAY ( title $ x121Address $ registeredAddress $ destinationIndicator $ preferredDeliveryMethod $ telexNumber $ teletexTerminalIdentifier $ telephoneNumber $ internationalISDNNumber $ facsimileTelephoneNumber $ street $ postOfficeBox $ postalCode $ postalAddress $ physicalDeliveryOfficeName $ ou $ st $ l ) )",
	"( 2.5.6.8 NAME 'organizationalRole' SUP top STRUCTURAL MUST cn MAY ( x121Address $ registeredAddress $ destinationIndicator $ preferredDeliveryMethod $ telexNumber $ teletexTerminalIdentifier $ telephoneNumber $ internationalISDNNumber $ facsimileTelephoneNumber $ seeAlso $ roleOccupant $ preferredDeliveryMethod $ street $ postOfficeBox $ postalCode $ postalAddress $ physicalDeliveryOfficeName $ ou $ st $ l $ description ) )",
	"( 2.5.6.5 NAME 'organizationalUnit' SUP top STRUCTURAL MUST ou MAY ( businessCategory $ description $ destinationIndicator $ facsimileTelephoneNumber $ internationalISDNNumber $ l $ physicalDeliveryOfficeName $ postalAddress $ postalCode $ postOfficeBox $ preferredDeliveryMethod $ registeredAddress $ searchGuide $ seeAlso $ st $ street $ telephoneNumber $ teletexTerminalIdentifier $ telexNumber $ userPassword $ x121Address ) )",
	"( 2.5.6.6 NAME 'person' SUP top STRUCTURAL MUST ( sn $ cn ) MAY ( userPassword $ telephoneNumber $ seeAlso $ description ) )",
	"( 2.5.6.10 NAME 'residentialPerson' SUP person STRUCTURAL MUST l MAY ( businessCategory $ x121Address $ registeredAddress $ destinationIndicator $ preferredDeliveryMethod $ telexNumber $ teletexTerminalIdentifier $ telephoneNumber $ internationalISDNNumber $ facsimileTelephoneNumber $ preferredDeliveryMethod $ street $ postOfficeBox $ postalCode $ postalAddress $ physicalDeliveryOfficeName $ st $ l ) )",
	"( 1.3.6.1.1.3.1 NAME 'uidObject' SUP top AUXILIARY MUST uid )",
)

RFC4519_SUBSCHEMA = directory.Subschema(ldapsyntaxes=CORE_SYNTAXES,
                                        matchingrules=CORE_MATCHING_RULES,
                                        attributetypes=RFC4519_ATTRIBUTE_TYPES,
                                        objectclasses=RFC4519_OBJECT_CLASSES)

# RFC 4524: COSINE LDAP/X.500 Schema

RFC4524_ATTRIBUTE_TYPES = RFC4519_ATTRIBUTE_TYPES + (
	"( 0.9.2342.19200300.100.1.37 NAME 'associatedDomain' EQUALITY caseIgnoreIA5Match SUBSTR caseIgnoreIA5SubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.26 )",
	"( 0.9.2342.19200300.100.1.38 NAME 'associatedName' EQUALITY distinguishedNameMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.12 )",
	"( 0.9.2342.19200300.100.1.48 NAME 'buildingName' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15{256} )",
	"( 0.9.2342.19200300.100.1.43 NAME 'co' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 )",
	"( 0.9.2342.19200300.100.1.14 NAME 'documentAuthor' EQUALITY distinguishedNameMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.12 )",
	"( 0.9.2342.19200300.100.1.11 NAME 'documentIdentifier' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15{256} )",
	"( 0.9.2342.19200300.100.1.15 NAME 'documentLocation' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15{256} )",
	"( 0.9.2342.19200300.100.1.56 NAME 'documentPublisher' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 )",
	"( 0.9.2342.19200300.100.1.12 NAME 'documentTitle' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15{256} )",
	"( 0.9.2342.19200300.100.1.13 NAME 'documentVersion' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15{256} )",
	"( 0.9.2342.19200300.100.1.5 NAME 'drink' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15{256} )",
	"( 0.9.2342.19200300.100.1.20 NAME 'homePhone' EQUALITY telephoneNumberMatch SUBSTR telephoneNumberSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.50 )",
	"( 0.9.2342.19200300.100.1.39 NAME 'homePostalAddress' EQUALITY caseIgnoreListMatch SUBSTR caseIgnoreListSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.41 )",
	"( 0.9.2342.19200300.100.1.9 NAME 'host' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15{256} )",
	"( 0.9.2342.19200300.100.1.4 NAME 'info' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15{2048} )",
	"( 0.9.2342.19200300.100.1.3 NAME 'mail' EQUALITY caseIgnoreIA5Match SUBSTR caseIgnoreIA5SubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.26{256} )",
	"( 0.9.2342.19200300.100.1.10 NAME 'manager' EQUALITY distinguishedNameMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.12 )",
	"( 0.9.2342.19200300.100.1.41 NAME 'mobile' EQUALITY telephoneNumberMatch SUBSTR telephoneNumberSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.50 )",
	"( 0.9.2342.19200300.100.1.45 NAME 'organizationalStatus' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15{256} )",
	"( 0.9.2342.19200300.100.1.42 NAME 'pager' EQUALITY telephoneNumberMatch SUBSTR telephoneNumberSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.50 )",
	"( 0.9.2342.19200300.100.1.40 NAME 'personalTitle' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15{256} )",
	"( 0.9.2342.19200300.100.1.6 NAME 'roomNumber' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15{256} )",
	"( 0.9.2342.19200300.100.1.21 NAME 'secretary' EQUALITY distinguishedNameMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.12 )",
	"( 0.9.2342.19200300.100.1.44 NAME 'uniqueIdentifier' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15{256} )",
	"( 0.9.2342.19200300.100.1.8 NAME 'userClass' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15{256} )",
)

RFC4524_OBJECT_CLASSES = RFC4519_OBJECT_CLASSES + (
	"( 0.9.2342.19200300.100.4.5 NAME 'account' SUP top STRUCTURAL MUST uid MAY ( description $ seeAlso $ l $ o $ ou $ host ) )",
	"( 0.9.2342.19200300.100.4.6 NAME 'document' SUP top STRUCTURAL MUST documentIdentifier MAY ( cn $ description $ seeAlso $ l $ o $ ou $ documentTitle $ documentVersion $ documentAuthor $ documentLocation $ documentPublisher ) )",
	"( 0.9.2342.19200300.100.4.9 NAME 'documentSeries' SUP top STRUCTURAL MUST cn MAY ( description $ l $ o $ ou $ seeAlso $ telephonenumber ) )",
	"( 0.9.2342.19200300.100.4.13 NAME 'domain' SUP top STRUCTURAL MUST dc MAY ( userPassword $ searchGuide $ seeAlso $ businessCategory $ x121Address $ registeredAddress $ destinationIndicator $ preferredDeliveryMethod $ telexNumber $ teletexTerminalIdentifier $ telephoneNumber $ internationaliSDNNumber $ facsimileTelephoneNumber $ street $ postOfficeBox $ postalCode $ postalAddress $ physicalDeliveryOfficeName $ st $ l $ description $ o $ associatedName ) )",
	"( 0.9.2342.19200300.100.4.17 NAME 'domainRelatedObject' SUP top AUXILIARY MUST associatedDomain )",
	"( 0.9.2342.19200300.100.4.18 NAME 'friendlyCountry' SUP country STRUCTURAL MUST co )",
	"( 0.9.2342.19200300.100.4.14 NAME 'rFC822localPart' SUP domain STRUCTURAL MAY ( cn $ description $ destinationIndicator $ facsimileTelephoneNumber $ internationaliSDNNumber $ physicalDeliveryOfficeName $ postalAddress $ postalCode $ postOfficeBox $ preferredDeliveryMethod $ registeredAddress $ seeAlso $ sn $ street $ telephoneNumber $ teletexTerminalIdentifier $ telexNumber $ x121Address ) )",
	"( 0.9.2342.19200300.100.4.7 NAME 'room' SUP top STRUCTURAL MUST cn MAY ( roomNumber $ description $ seeAlso $ telephoneNumber ) )",
	"( 0.9.2342.19200300.100.4.19 NAME 'simpleSecurityObject' SUP top AUXILIARY MUST userPassword )",
)

RFC4524_SUBSCHEMA = directory.Subschema(ldapsyntaxes=CORE_SYNTAXES,
                                        matchingrules=CORE_MATCHING_RULES,
                                        attributetypes=RFC4524_ATTRIBUTE_TYPES,
                                        objectclasses=RFC4524_OBJECT_CLASSES)
COSINE_SUBSCHEMA = RFC4524_SUBSCHEMA

# RFC 2307: NIS Schema

RFC2307_ATTRIBUTE_TYPES = RFC4519_ATTRIBUTE_TYPES + (
	"( 1.3.6.1.1.1.1.0 NAME 'uidNumber' DESC 'An integer uniquely identifying a user in an administrative domain' EQUALITY integerMatch SYNTAX 'INTEGER' SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.1 NAME 'gidNumber' DESC 'An integer uniquely identifying a group in an administrative domain' EQUALITY integerMatch SYNTAX 'INTEGER' SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.2 NAME 'gecos' DESC 'The GECOS field; the common name' EQUALITY caseIgnoreIA5Match SUBSTRINGS caseIgnoreIA5SubstringsMatch SYNTAX 'IA5String' SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.3 NAME 'homeDirectory' DESC 'The absolute path to the home directory' EQUALITY caseExactIA5Match SYNTAX 'IA5String' SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.4 NAME 'loginShell' DESC 'The path to the login shell' EQUALITY caseExactIA5Match SYNTAX 'IA5String' SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.5 NAME 'shadowLastChange' EQUALITY integerMatch SYNTAX 'INTEGER' SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.6 NAME 'shadowMin' EQUALITY integerMatch SYNTAX 'INTEGER' SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.7 NAME 'shadowMax' EQUALITY integerMatch SYNTAX 'INTEGER' SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.8 NAME 'shadowWarning' EQUALITY integerMatch SYNTAX 'INTEGER' SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.9 NAME 'shadowInactive' EQUALITY integerMatch SYNTAX 'INTEGER' SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.10 NAME 'shadowExpire' EQUALITY integerMatch SYNTAX 'INTEGER' SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.11 NAME 'shadowFlag' EQUALITY integerMatch SYNTAX 'INTEGER' SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.12 NAME 'memberUid' EQUALITY caseExactIA5Match SUBSTRINGS caseExactIA5SubstringsMatch SYNTAX 'IA5String' )",
	"( 1.3.6.1.1.1.1.13 NAME 'memberNisNetgroup' EQUALITY caseExactIA5Match SUBSTRINGS caseExactIA5SubstringsMatch SYNTAX 'IA5String' )",
	"( 1.3.6.1.1.1.1.14 NAME 'nisNetgroupTriple' DESC 'Netgroup triple' SYNTAX 'nisNetgroupTripleSyntax' ) ( 1.3.6.1.1.1.1.15 NAME 'ipServicePort' EQUALITY integerMatch SYNTAX 'INTEGER' SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.16 NAME 'ipServiceProtocol' SUP name ) ( 1.3.6.1.1.1.1.17 NAME 'ipProtocolNumber' EQUALITY integerMatch SYNTAX 'INTEGER' SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.18 NAME 'oncRpcNumber' EQUALITY integerMatch SYNTAX 'INTEGER' SINGLE-VALUE ) ( 1.3.6.1.1.1.1.19 NAME 'ipHostNumber' DESC 'IP address as a dotted decimal, eg. 192.168.1.1, omitting leading zeros' EQUALITY caseIgnoreIA5Match SYNTAX 'IA5String{128}' )",
	"( 1.3.6.1.1.1.1.20 NAME 'ipNetworkNumber' DESC 'IP network as a dotted decimal, eg. 192.168, omitting leading zeros' EQUALITY caseIgnoreIA5Match SYNTAX 'IA5String{128}' SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.21 NAME 'ipNetmaskNumber' DESC 'IP netmask as a dotted decimal, eg. 255.255.255.0, omitting leading zeros' EQUALITY caseIgnoreIA5Match SYNTAX 'IA5String{128}' SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.22 NAME 'macAddress' DESC 'MAC address in maximal, colon separated hex notation, eg. 00:00:92:90:ee:e2' EQUALITY caseIgnoreIA5Match SYNTAX 'IA5String{128}' )",
	"( 1.3.6.1.1.1.1.23 NAME 'bootParameter' DESC 'rpc.bootparamd parameter' SYNTAX 'bootParameterSyntax' )",
	"( 1.3.6.1.1.1.1.24 NAME 'bootFile' DESC 'Boot image name' EQUALITY caseExactIA5Match SYNTAX 'IA5String' )",
	"( 1.3.6.1.1.1.1.26 NAME 'nisMapName' SUP name )",
	"( 1.3.6.1.1.1.1.27 NAME 'nisMapEntry' EQUALITY caseExactIA5Match SUBSTRINGS caseExactIA5SubstringsMatch SYNTAX 'IA5String{1024}' SINGLE-VALUE )",
)

RFC2307_OBJECT_CLASSES = RFC4519_OBJECT_CLASSES + (
	"( 1.3.6.1.1.1.2.0 NAME 'posixAccount' SUP top AUXILIARY DESC 'Abstraction of an account with POSIX attributes' MUST ( cn $ uid $ uidNumber $ gidNumber $ homeDirectory ) MAY ( userPassword $ loginShell $ gecos $ description ) )",
	"( 1.3.6.1.1.1.2.1 NAME 'shadowAccount' SUP top AUXILIARY DESC 'Additional attributes for shadow passwords' MUST uid MAY ( userPassword $ shadowLastChange $ shadowMin $ shadowMax $ shadowWarning $ shadowInactive $ shadowExpire $ shadowFlag $ description ) )",
	"( 1.3.6.1.1.1.2.2 NAME 'posixGroup' SUP top STRUCTURAL DESC 'Abstraction of a group of accounts' MUST ( cn $ gidNumber ) MAY ( userPassword $ memberUid $ description ) )",
	"( 1.3.6.1.1.1.2.3 NAME 'ipService' SUP top STRUCTURAL DESC 'Abstraction an Internet Protocol service.  Maps an IP port and protocol (such as tcp or udp) to one or more names; the distinguished value of the cn attribute denotes the service's canonical name' MUST ( cn $ ipServicePort $ ipServiceProtocol ) MAY ( description ) )",
	"( 1.3.6.1.1.1.2.4 NAME 'ipProtocol' SUP top STRUCTURAL DESC 'Abstraction of an IP protocol. Maps a protocol number to one or more names. The distinguished value of the cn attribute denotes the protocol's canonical name' MUST ( cn $ ipProtocolNumber $ description ) MAY description )",
	"( 1.3.6.1.1.1.2.5 NAME 'oncRpc' SUP top STRUCTURAL DESC 'Abstraction of an Open Network Computing (ONC) [RFC1057] Remote Procedure Call (RPC) binding.  This class maps an ONC RPC number to a name.  The distinguished value of the cn attribute denotes the RPC service's canonical name' MUST ( cn $ oncRpcNumber $ description ) MAY description )",
	"( 1.3.6.1.1.1.2.6 NAME 'ipHost' SUP top AUXILIARY DESC 'Abstraction of a host, an IP device. The distinguished value of the cn attribute denotes the host's canonical name. Device SHOULD be used as a structural class' MUST ( cn $ ipHostNumber ) MAY ( l $ description $ manager ) )",
	"( 1.3.6.1.1.1.2.7 NAME 'ipNetwork' SUP top STRUCTURAL DESC 'Abstraction of a network. The distinguished value of the cn attribute denotes the network's canonical name' MUST ( cn $ ipNetworkNumber ) MAY ( ipNetmaskNumber $ l $ description $ manager ) )",
	"( 1.3.6.1.1.1.2.8 NAME 'nisNetgroup' SUP top STRUCTURAL DESC 'Abstraction of a netgroup. May refer to other netgroups' MUST cn MAY ( nisNetgroupTriple $ memberNisNetgroup $ description ) )",
	"( 1.3.6.1.1.1.2.09 NAME 'nisMap' SUP top STRUCTURAL DESC 'A generic abstraction of a NIS map' MUST nisMapName MAY description )",
	"( 1.3.6.1.1.1.2.10 NAME 'nisObject' SUP top STRUCTURAL DESC 'An entry in a NIS map' MUST ( cn $ nisMapEntry $ nisMapName ) MAY description )",
	"( 1.3.6.1.1.1.2.11 NAME 'ieee802Device' SUP top AUXILIARY DESC 'A device with a MAC address; device SHOULD be used as a structural class' MAY macAddress )",
	"( 1.3.6.1.1.1.2.12 NAME 'bootableDevice' SUP top AUXILIARY DESC 'A device with boot parameters; device SHOULD be used as a structural class' MAY ( bootFile $ bootParameter ) )",
)

RFC2307_SUBSCHEMA = directory.Subschema(ldapsyntaxes=CORE_SYNTAXES,
                                        matchingrules=CORE_MATCHING_RULES,
                                        attributetypes=RFC2307_ATTRIBUTE_TYPES,
                                        objectclasses=RFC2307_OBJECT_CLASSES)
NIS_SUBSCHEMA = RFC2307_SUBSCHEMA

RFC2307BIS_ATTRIBUTE_TYPES = RFC4519_ATTRIBUTE_TYPES + (
	"( 1.3.6.1.1.1.1.0 NAME 'uidNumber' DESC 'An integer uniquely identifying a user in an administrative domain' EQUALITY integerMatch ORDERING integerOrderingMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.27 SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.1 NAME 'gidNumber' DESC 'An integer uniquely identifying a group in an administrative domain' EQUALITY integerMatch ORDERING integerOrderingMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.27 SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.2 NAME 'gecos' DESC 'The GECOS field; the common name' EQUALITY caseIgnoreMatch SUBSTRINGS caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.3 NAME 'homeDirectory' DESC 'The absolute path to the home directory' EQUALITY caseExactIA5Match SYNTAX 1.3.6.1.4.1.1466.115.121.1.26 SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.4 NAME 'loginShell' DESC 'The path to the login shell' EQUALITY caseExactIA5Match SYNTAX 1.3.6.1.4.1.1466.115.121.1.26 SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.5 NAME 'shadowLastChange' EQUALITY integerMatch ORDERING integerOrderingMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.27 SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.6 NAME 'shadowMin' EQUALITY integerMatch ORDERING integerOrderingMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.27 SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.7 NAME 'shadowMax' EQUALITY integerMatch ORDERING integerOrderingMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.27 SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.8 NAME 'shadowWarning' EQUALITY integerMatch ORDERING integerOrderingMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.27 SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.9 NAME 'shadowInactive' EQUALITY integerMatch ORDERING integerOrderingMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.27 SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.10 NAME 'shadowExpire' EQUALITY integerMatch ORDERING integerOrderingMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.27 SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.11 NAME 'shadowFlag' EQUALITY integerMatch ORDERING integerOrderingMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.27 SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.12 NAME 'memberUid' EQUALITY caseExactMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 )",
	"( 1.3.6.1.1.1.1.13 NAME 'memberNisNetgroup' EQUALITY caseExactMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 )",
	"( 1.3.6.1.1.1.1.14 NAME 'nisNetgroupTriple' DESC 'Netgroup triple' EQUALITY caseIgnoreMatch SUBSTRINGS caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 )",
	"( 1.3.6.1.1.1.1.15 NAME 'ipServicePort' DESC 'Service port number' EQUALITY integerMatch ORDERING integerOrderingMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.27 SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.16 NAME 'ipServiceProtocol' DESC 'Service protocol name' EQUALITY caseIgnoreMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 )",
	"( 1.3.6.1.1.1.1.17 NAME 'ipProtocolNumber' DESC 'IP protocol number' EQUALITY integerMatch ORDERING integerOrderingMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.27 SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.18 NAME 'oncRpcNumber' DESC 'ONC RPC number' EQUALITY integerMatch ORDERING integerOrderingMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.27 SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.19 NAME 'ipHostNumber' DESC 'IPv4 addresses as a dotted decimal omitting leading zeros or IPv6 addresses as defined in RFC2373' EQUALITY caseIgnoreIA5Match SYNTAX 1.3.6.1.4.1.1466.115.121.1.26 )",
	"( 1.3.6.1.1.1.1.20 NAME 'ipNetworkNumber' DESC 'IP network omitting leading zeros, eg. 192.168' EQUALITY caseIgnoreIA5Match SYNTAX 1.3.6.1.4.1.1466.115.121.1.26 SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.21 NAME 'ipNetmaskNumber' DESC 'IP netmask omitting leading zeros, eg. 255.255.255.0' EQUALITY caseIgnoreIA5Match SYNTAX 1.3.6.1.4.1.1466.115.121.1.26 SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.22 NAME 'macAddress' DESC 'MAC address in maximal, colon separated hex notation, eg. 00:00:92:90:ee:e2' EQUALITY caseIgnoreIA5Match SYNTAX 1.3.6.1.4.1.1466.115.121.1.26 )",
	"( 1.3.6.1.1.1.1.23 NAME 'bootParameter' DESC 'rpc.bootparamd parameter' EQUALITY caseExactIA5Match SYNTAX 1.3.6.1.4.1.1466.115.121.1.26 )",
	"( 1.3.6.1.1.1.1.24 NAME 'bootFile' DESC 'Boot image name' EQUALITY caseExactIA5Match SYNTAX 1.3.6.1.4.1.1466.115.121.1.26 )",
	"( 1.3.6.1.1.1.1.26 NAME 'nisMapName' DESC 'Name of a generic NIS map' EQUALITY caseIgnoreMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15{64} )",
	"( 1.3.6.1.1.1.1.27 NAME 'nisMapEntry' DESC 'A generic NIS entry' EQUALITY caseExactMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15{1024} SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.28 NAME 'nisPublicKey' DESC 'NIS public key' EQUALITY octetStringMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.40 SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.29 NAME 'nisSecretKey' DESC 'NIS secret key' EQUALITY octetStringMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.40 SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.30 NAME 'nisDomain' DESC 'NIS domain' EQUALITY caseIgnoreIA5Match SYNTAX 1.3.6.1.4.1.1466.115.121.1.26{256} )",
	"( 1.3.6.1.1.1.1.31 NAME 'automountMapName' DESC 'automount Map Name' EQUALITY caseExactMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.32 NAME 'automountKey' DESC 'Automount Key value' EQUALITY caseExactMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.33 NAME 'automountInformation' DESC 'Automount information' EQUALITY caseExactMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 SINGLE-VALUE )",
)

RFC2307BIS_OBJECT_CLASSES = RFC4519_OBJECT_CLASSES + (
	"( 1.3.6.1.1.1.2.0 NAME 'posixAccount' SUP top AUXILIARY DESC 'Abstraction of an account with POSIX attributes' MUST ( cn $ uid $ uidNumber $ gidNumber $ homeDirectory ) MAY ( authPassword $ userPassword $ loginShell $ gecos $ description ) )",
	"( 1.3.6.1.1.1.2.1 NAME 'shadowAccount' SUP top AUXILIARY DESC 'Additional attributes for shadow passwords' MUST uid MAY ( authPassword $ userPassword $ description $ shadowLastChange $ shadowMin $ shadowMax $ shadowWarning $ shadowInactive $ shadowExpire $ shadowFlag ) )",
	"( 1.3.6.1.1.1.2.2 NAME 'posixGroup' SUP top AUXILIARY DESC 'Abstraction of a group of accounts' MUST gidNumber MAY ( authPassword $ userPassword $ memberUid $ description ) )",
	"( 1.3.6.1.1.1.2.3 NAME 'ipService' SUP top STRUCTURAL DESC 'Abstraction an Internet Protocol service.  Maps an IP port and protocol (such as tcp or udp) to one or more names; the distinguished value of the cn attribute denotes the service's canonical name' MUST ( cn $ ipServicePort $ ipServiceProtocol ) MAY description )",
	"( 1.3.6.1.1.1.2.4 NAME 'ipProtocol' SUP top STRUCTURAL DESC 'Abstraction of an IP protocol. Maps a protocol number to one or more names. The distinguished value of the cn attribute denotes the protocol canonical name' MUST ( cn $ ipProtocolNumber ) MAY description )",
	"( 1.3.6.1.1.1.2.5 NAME 'oncRpc' SUP top STRUCTURAL DESC 'Abstraction of an Open Network Computing (ONC) [RFC1057] Remote Procedure Call (RPC) binding.  This class maps an ONC RPC number to a name.  The distinguished value of the cn attribute denotes the RPC service canonical name' MUST ( cn $ oncRpcNumber ) MAY description )",
	"( 1.3.6.1.1.1.2.6 NAME 'ipHost' SUP top AUXILIARY DESC 'Abstraction of a host, an IP device. The distinguished value of the cn attribute denotes the host's canonical name. Device SHOULD be used as a structural class' MUST ( cn $ ipHostNumber ) MAY ( authPassword $ userPassword $ l $ description $ manager ) )",
	"( 1.3.6.1.1.1.2.7 NAME 'ipNetwork' SUP top STRUCTURAL DESC 'Abstraction of a network. The distinguished value of the cn attribute denotes the network canonical name' MUST ipNetworkNumber MAY ( cn $ ipNetmaskNumber $ l $ description $ manager ) )",
	"( 1.3.6.1.1.1.2.8 NAME 'nisNetgroup' SUP top STRUCTURAL DESC 'Abstraction of a netgroup. May refer to other netgroups' MUST cn MAY ( nisNetgroupTriple $ memberNisNetgroup $ description ) )",
	"( 1.3.6.1.1.1.2.9 NAME 'nisMap' SUP top STRUCTURAL DESC 'A generic abstraction of a NIS map' MUST nisMapName MAY description )",
	"( 1.3.6.1.1.1.2.10 NAME 'nisObject' SUP top STRUCTURAL DESC 'An entry in a NIS map' MUST ( cn $ nisMapEntry $ nisMapName )",
	"( 1.3.6.1.1.1.2.11 NAME 'ieee802Device' SUP top AUXILIARY DESC 'A device with a MAC address; device SHOULD be used as a structural class' MAY macAddress )",
	"( 1.3.6.1.1.1.2.12 NAME 'bootableDevice' SUP top AUXILIARY DESC 'A device with boot parameters; device SHOULD be used as a structural class' MAY ( bootFile $ bootParameter ) )",
	"( 1.3.6.1.1.1.2.14 NAME 'nisKeyObject' SUP top AUXILIARY DESC 'An object with a public and secret key' MUST ( cn $ nisPublicKey $ nisSecretKey ) MAY ( uidNumber $ description ) )",
	"( 1.3.6.1.1.1.2.15 NAME 'nisDomainObject' SUP top AUXILIARY DESC 'Associates a NIS domain with a naming context' MUST nisDomain )",
	"( 1.3.6.1.1.1.2.16 NAME 'automountMap' SUP top STRUCTURAL MUST ( automountMapName ) MAY description )",
	"( 1.3.6.1.1.1.2.17 NAME 'automount' SUP top STRUCTURAL DESC 'Automount information' MUST ( automountKey $ automountInformation ) MAY description )",
	"( 1.3.6.1.1.1.2.18 NAME 'groupOfMembers' SUP top STRUCTURAL DESC 'A group with members (DNs)' MUST cn MAY ( businessCategory $ seeAlso $ owner $ ou $ o $ description $ member ) )",
)

RFC2307BIS_SUBSCHEMA = directory.Subschema(ldapsyntaxes=CORE_SYNTAXES,
                                           matchingrules=CORE_MATCHING_RULES,
                                           attributetypes=RFC2307BIS_ATTRIBUTE_TYPES,
                                           objectclasses=RFC2307BIS_OBJECT_CLASSES)
